#include <mesh/StencilManager.hpp>

#include <mesh/StencilBuilder.hpp>
#include <utils/Exceptions.hpp>

StencilManager* StencilManager::m_instance = nullptr;

void
StencilManager::create()
{
  Assert(m_instance == nullptr, "StencilManager is already created");
  m_instance = new StencilManager;
}

void
StencilManager::destroy()
{
  Assert(m_instance != nullptr, "StencilManager was not created");

  // LCOV_EXCL_START
  auto check_still_registered = [](auto& stored_stencil_map) {
    if (stored_stencil_map.size() > 0) {
      std::stringstream error;
      error << ": some connectivities are still registered\n";
      for (const auto& [key, stencil_p] : stored_stencil_map) {
        error << " - connectivity of id " << rang::fgB::magenta << key.connectivity_id << rang::style::reset << '\n';
      }
      throw UnexpectedError(error.str());
    }
  };

  check_still_registered(m_instance->m_stored_cell_to_cell_stencil_map);
  check_still_registered(m_instance->m_stored_node_to_cell_stencil_map);
  // LCOV_EXCL_STOP

  delete m_instance;
  m_instance = nullptr;
}

template <ItemType source_item_type, ItemType target_item_type>
const StencilArray<source_item_type, target_item_type>&
StencilManager::_getStencilArray(
  const IConnectivity& connectivity,
  const StencilDescriptor& stencil_descriptor,
  const BoundaryDescriptorList& symmetry_boundary_descriptor_list,
  StoredStencilTMap<source_item_type, target_item_type>& stored_source_to_target_stencil_map)
{
  if (not stored_source_to_target_stencil_map.contains(
        Key{connectivity.id(), stencil_descriptor, symmetry_boundary_descriptor_list})) {
    stored_source_to_target_stencil_map[Key{connectivity.id(), stencil_descriptor, symmetry_boundary_descriptor_list}] =
      std::make_shared<StencilArray<source_item_type, target_item_type>>(
        StencilBuilder{}.template _build<source_item_type, target_item_type>(connectivity, stencil_descriptor,
                                                                             symmetry_boundary_descriptor_list));
  }

  return *stored_source_to_target_stencil_map.at(
    Key{connectivity.id(), stencil_descriptor, symmetry_boundary_descriptor_list});
}

const CellToCellStencilArray&
StencilManager::getCellToCellStencilArray(const IConnectivity& connectivity,
                                          const StencilDescriptor& stencil_descriptor,
                                          const BoundaryDescriptorList& symmetry_boundary_descriptor_list)
{
  return this->_getStencilArray<ItemType::cell, ItemType::cell>(connectivity, stencil_descriptor,
                                                                symmetry_boundary_descriptor_list,
                                                                m_stored_cell_to_cell_stencil_map);
}

const NodeToCellStencilArray&
StencilManager::getNodeToCellStencilArray(const IConnectivity& connectivity,
                                          const StencilDescriptor& stencil_descriptor,
                                          const BoundaryDescriptorList& symmetry_boundary_descriptor_list)
{
  return this->_getStencilArray<ItemType::node, ItemType::cell>(connectivity, stencil_descriptor,
                                                                symmetry_boundary_descriptor_list,
                                                                m_stored_node_to_cell_stencil_map);
}

void
StencilManager::deleteConnectivity(const size_t connectivity_id)
{
  auto delete_connectivity_stencil = [&connectivity_id](auto& stored_stencil_map) {
    bool has_removed = false;
    do {
      has_removed = false;
      for (const auto& [key, p_stencil] : stored_stencil_map) {
        if (connectivity_id == key.connectivity_id) {
          stored_stencil_map.erase(key);
          has_removed = true;
          break;
        }
      }
    } while (has_removed);
  };

  delete_connectivity_stencil(m_stored_cell_to_cell_stencil_map);
  delete_connectivity_stencil(m_stored_node_to_cell_stencil_map);
}
