#include <catch2/catch_test_macros.hpp>
#include <catch2/matchers/catch_matchers_all.hpp>

#include <MeshDataBaseForTests.hpp>
#include <mesh/Mesh.hpp>
#include <scheme/DiscreteFunctionDPk.hpp>
#include <scheme/DiscreteFunctionP0.hpp>

// clazy:excludeall=non-pod-global-static

TEST_CASE("DiscreteFunctionDPk", "[scheme]")
{
  SECTION("Basic interface")
  {
    const std::shared_ptr mesh_2d_v = MeshDataBaseForTests::get().cartesian2DMesh();
    const std::shared_ptr mesh_2d   = mesh_2d_v->get<Mesh<2>>();

    const std::shared_ptr mesh_2d_2_v = MeshDataBaseForTests::get().hybrid2DMesh();
    const std::shared_ptr mesh_2d_2   = mesh_2d_2_v->get<Mesh<2>>();

    DiscreteFunctionDPk<2, const double> R_dkp;

#ifndef NDEBUG
    REQUIRE_THROWS_WITH(R_dkp[CellId(0)], "DiscreteFunctionDPk is not built");
#endif   // NDEBUG

    {
      DiscreteFunctionDPk<2, double> tmp_R_dkp(mesh_2d_v, 2);
      tmp_R_dkp.fill(2);
      R_dkp = tmp_R_dkp;
    }

    REQUIRE(R_dkp.degree() == 2);

    REQUIRE(min(R_dkp.cellArrays()) == 2);
    REQUIRE(max(R_dkp.cellArrays()) == 2);

    REQUIRE(R_dkp.dataType() == ast_node_data_type_from<double>);
    REQUIRE(R_dkp.meshVariant()->id() == mesh_2d->id());

    DiscreteFunctionDPk<2, double> R_dkp2;
    R_dkp2 = copy(R_dkp);

    REQUIRE(R_dkp2.degree() == 2);

    REQUIRE(min(R_dkp2.cellArrays()) == 2);
    REQUIRE(max(R_dkp2.cellArrays()) == 2);

    REQUIRE(R_dkp2.dataType() == ast_node_data_type_from<double>);
    REQUIRE(R_dkp2.meshVariant()->id() == mesh_2d_v->id());

    DiscreteFunctionDPk<2, double> R_dkp3(mesh_2d_2_v, 1);
    R_dkp3.fill(3);

    REQUIRE(R_dkp3.degree() == 1);

    REQUIRE(min(R_dkp3.cellArrays()) == 3);
    REQUIRE(max(R_dkp3.cellArrays()) == 3);

    REQUIRE(R_dkp3.dataType() == ast_node_data_type_from<double>);
    REQUIRE(R_dkp3.meshVariant()->id() == mesh_2d_2_v->id());

    DiscreteFunctionDPk<2, double> R_dkp4(mesh_2d, R_dkp2.cellArrays());

    REQUIRE(min(R_dkp4.cellArrays()) == 2);
    REQUIRE(max(R_dkp4.cellArrays()) == 2);

    R_dkp4.fill(5);
    REQUIRE(min(R_dkp4.cellArrays()) == 5);
    REQUIRE(max(R_dkp4.cellArrays()) == 5);
    REQUIRE(min(R_dkp2.cellArrays()) == 5);
    REQUIRE(max(R_dkp2.cellArrays()) == 5);

    copy_to(R_dkp, R_dkp4);
    REQUIRE(min(R_dkp4.cellArrays()) == 2);
    REQUIRE(max(R_dkp4.cellArrays()) == 2);
    REQUIRE(min(R_dkp2.cellArrays()) == 2);
    REQUIRE(max(R_dkp2.cellArrays()) == 2);

#ifndef NDEBUG
    REQUIRE_THROWS_WITH(copy_to(R_dkp, R_dkp3), "copy_to target must use the same mesh");

    DiscreteFunctionDPk<2, double> R_dkp5(mesh_2d, 3);
    REQUIRE_THROWS_WITH(copy_to(R_dkp, R_dkp5), "copy_to target must have the same degree");

    REQUIRE_THROWS_WITH((DiscreteFunctionDPk<2, double>{mesh_2d_2, R_dkp2.cellArrays()}),
                        "cell_array is built on different connectivity");
#endif   // NDEBUG
  }

  SECTION("R data")
  {
    SECTION("1D")
    {
      constexpr size_t Dimension = 1;

      using R1 = TinyVector<1>;

      const std::shared_ptr mesh_v = MeshDataBaseForTests::get().cartesian1DMesh();
      const std::shared_ptr mesh   = mesh_v->get<Mesh<Dimension>>();

      auto xj = MeshDataManager::instance().getMeshData(*mesh).xj();
      auto Vj = MeshDataManager::instance().getMeshData(*mesh).Vj();

      const size_t degree = 4;

      DiscreteFunctionDPk<Dimension, double> pk(mesh_v, degree);

      std::vector<double> a = {1, 1.4, -6.2, 2.7, 3.1};

      for (CellId cell_id = 0; cell_id < mesh->numberOfCells(); ++cell_id) {
        auto coefficients = pk.coefficients(cell_id);
        for (size_t i = 0; i < a.size(); ++i) {
          coefficients[i] = a[i];
        }
      };

      DiscreteFunctionP0<double> p_xj(mesh_v);
      parallel_for(
        mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) { p_xj[cell_id] = pk[cell_id](xj[cell_id]); });

      REQUIRE(max(p_xj) == 1);
      REQUIRE(min(p_xj) == 1);

      DiscreteFunctionP0<double> delta(mesh_v);
      parallel_for(
        mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) {
          const double x0 = 0.3 * Vj[cell_id];
          delta[cell_id]  = pk[cell_id](xj[cell_id] + R1{x0})   //
                           - (a[0] + x0 * a[1] + x0 * x0 * a[2] + x0 * x0 * x0 * a[3] + x0 * x0 * x0 * x0 * a[4]);
        });

      REQUIRE(max(delta) == Catch::Approx(0.).margin(1E-14));
      REQUIRE(min(delta) == Catch::Approx(0.).margin(1E-14));
    }

    SECTION("2D")
    {
      constexpr size_t Dimension = 2;

      using R2 = TinyVector<2>;

      const std::shared_ptr mesh_v = MeshDataBaseForTests::get().cartesian2DMesh();
      const std::shared_ptr mesh   = mesh_v->get<Mesh<Dimension>>();

      auto xj = MeshDataManager::instance().getMeshData(*mesh).xj();
      auto Vj = MeshDataManager::instance().getMeshData(*mesh).Vj();

      const size_t degree = 3;

      DiscreteFunctionDPk<Dimension, double> pk(mesh_v, degree);

      const std::vector<double> a = {1, 1.4, -6.2, 3.5, -2.3, 5.2, 6.1, 2.3, 0.5, -1.3};

      for (CellId cell_id = 0; cell_id < mesh->numberOfCells(); ++cell_id) {
        auto coefficients = pk.coefficients(cell_id);
        for (size_t i = 0; i < coefficients.size(); ++i) {
          coefficients[i] = a[i];
        }
      };

      DiscreteFunctionP0<double> p_xj(mesh_v);
      parallel_for(
        mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) { p_xj[cell_id] = pk[cell_id](xj[cell_id]); });

      REQUIRE(max(p_xj) == 1);
      REQUIRE(min(p_xj) == 1);

      DiscreteFunctionP0<double> error(mesh_v);
      parallel_for(
        mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) {
          const double x0 = 0.2 * Vj[cell_id];
          const double y0 = 0.3 * Vj[cell_id];
          error[cell_id]                                       //
            = std::abs(pk[cell_id](xj[cell_id] + R2{x0, y0})   //
                       - (a[0]                                 //
                          + x0 * a[1]                          //
                          + x0 * x0 * a[2]                     //
                          + x0 * x0 * x0 * a[3]                //
                          + y0 * a[4]                          //
                          + y0 * x0 * a[5]                     //
                          + y0 * x0 * x0 * a[6]                //
                          + y0 * y0 * a[7]                     //
                          + y0 * y0 * x0 * a[8]                //
                          + y0 * y0 * y0 * a[9]));
        });

      REQUIRE(max(error) == Catch::Approx(0.).margin(1E-14));
    }

    SECTION("3D")
    {
      constexpr size_t Dimension = 3;

      using R3 = TinyVector<3>;

      const std::shared_ptr mesh_v = MeshDataBaseForTests::get().cartesian3DMesh();
      const std::shared_ptr mesh   = mesh_v->get<Mesh<Dimension>>();

      auto xj = MeshDataManager::instance().getMeshData(*mesh).xj();
      auto Vj = MeshDataManager::instance().getMeshData(*mesh).Vj();

      const size_t degree = 3;

      DiscreteFunctionDPk<Dimension, double> pk(mesh_v, degree);

      const std::vector<double> a = {+1.0, +1.4, -6.2, +3.5, -2.3, +5.2, +6.1, +2.3, +0.5, -1.3,   //
                                     +2.8, -8.4, +9.5, +4.0, +4.3, +7.2, -9.1, +6.8, +6.7, +9.2};

      parallel_for(
        mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) {
          auto coefficients = pk.coefficients(cell_id);
          for (size_t i = 0; i < coefficients.size(); ++i) {
            coefficients[i] = a[i];
          }
        });

      DiscreteFunctionP0<double> delta_xj(mesh_v);
      parallel_for(
        mesh->numberOfCells(),
        PUGS_LAMBDA(CellId cell_id) { delta_xj[cell_id] = std::abs(pk[cell_id](xj[cell_id]) - a[0]); });

      REQUIRE(max(delta_xj) == Catch::Approx(0.).margin(1E-14));

      DiscreteFunctionP0<double> delta(mesh_v);
      parallel_for(
        mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) {
          const double x0 = +0.5 * Vj[cell_id];
          const double y0 = +0.3 * Vj[cell_id];
          const double z0 = -0.4 * Vj[cell_id];

          delta[cell_id] = pk[cell_id](xj[cell_id] + R3{x0, y0, z0})   //
                           - (a[0]                                     //
                              + x0 * a[1]                              //
                              + x0 * x0 * a[2]                         //
                              + x0 * x0 * x0 * a[3]                    //
                              + y0 * a[4]                              //
                              + y0 * x0 * a[5]                         //
                              + y0 * x0 * x0 * a[6]                    //
                              + y0 * y0 * a[7]                         //
                              + y0 * y0 * x0 * a[8]                    //
                              + y0 * y0 * y0 * a[9]                    //
                              + z0 * a[10]                             //
                              + z0 * x0 * a[11]                        //
                              + z0 * x0 * x0 * a[12]                   //
                              + z0 * y0 * a[13]                        //
                              + z0 * y0 * x0 * a[14]                   //
                              + z0 * y0 * y0 * a[15]                   //
                              + z0 * z0 * a[16]                        //
                              + z0 * z0 * x0 * a[17]                   //
                              + z0 * z0 * y0 * a[18]                   //
                              + z0 * z0 * z0 * a[19]                   //
                             );
        });

      REQUIRE(max(delta) == Catch::Approx(0.).margin(1E-14));
      REQUIRE(min(delta) == Catch::Approx(0.).margin(1E-14));
    }
  }

  SECTION("R^d data")
  {
    SECTION("1D")
    {
      constexpr size_t Dimension = 1;

      using R1 = TinyVector<1>;
      using R2 = TinyVector<2>;

      const std::shared_ptr mesh_v = MeshDataBaseForTests::get().cartesian1DMesh();
      const std::shared_ptr mesh   = mesh_v->get<Mesh<Dimension>>();

      auto xj = MeshDataManager::instance().getMeshData(*mesh).xj();
      auto Vj = MeshDataManager::instance().getMeshData(*mesh).Vj();

      const size_t degree = 4;

      DiscreteFunctionDPk<Dimension, R2> pk(mesh_v, degree);

      const std::vector<R2> a = {R2{-1.0, +3.0}, R2{+1.4, +1.9}, R2{-6.2, -1.0}, R2{+2.7, +1.6}, R2{+3.1, -1.3}};

      for (CellId cell_id = 0; cell_id < mesh->numberOfCells(); ++cell_id) {
        auto coefficients = pk.coefficients(cell_id);
        for (size_t i = 0; i < a.size(); ++i) {
          coefficients[i] = a[i];
        }
      }

      DiscreteFunctionP0<double> delta_xj(mesh_v);
      parallel_for(
        mesh->numberOfCells(),
        PUGS_LAMBDA(CellId cell_id) { delta_xj[cell_id] = l2Norm(pk[cell_id](xj[cell_id]) - a[0]); });

      REQUIRE(max(delta_xj) == Catch::Approx(0).margin(1E-14));

      DiscreteFunctionP0<double> delta(mesh_v);
      parallel_for(
        mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) {
          const double x0 = 0.3 * Vj[cell_id];
          delta[cell_id] =
            l2Norm(pk[cell_id](xj[cell_id] + R1{x0})   //
                   - (a[0] + x0 * a[1] + x0 * x0 * a[2] + x0 * x0 * x0 * a[3] + x0 * x0 * x0 * x0 * a[4]));
        });

      REQUIRE(max(delta) == Catch::Approx(0.).margin(1E-14));
    }

    SECTION("2D")
    {
      constexpr size_t Dimension = 2;

      using R2 = TinyVector<2>;

      const std::shared_ptr mesh_v = MeshDataBaseForTests::get().hybrid2DMesh();
      const std::shared_ptr mesh   = mesh_v->get<Mesh<Dimension>>();

      auto xj = MeshDataManager::instance().getMeshData(*mesh).xj();
      auto Vj = MeshDataManager::instance().getMeshData(*mesh).Vj();

      const size_t degree = 3;

      DiscreteFunctionDPk<Dimension, R2> pk(mesh_v, degree);

      const std::vector<R2> a = {R2{-1.0, +3.0}, R2{+1.4, +1.9}, R2{-6.2, -1.0}, R2{+2.7, +1.6}, R2{+3.1, -1.3},
                                 R2{+2.5, -4.2}, R2{+2.1, -1.7}, R2{-3.2, +1.0}, R2{-2.3, +1.3}, R2{-2.9, -3.2}};

      for (CellId cell_id = 0; cell_id < mesh->numberOfCells(); ++cell_id) {
        auto coefficients = pk.coefficients(cell_id);
        for (size_t i = 0; i < a.size(); ++i) {
          coefficients[i] = a[i];
        }
      }

      DiscreteFunctionP0<double> delta_xj(mesh_v);
      parallel_for(
        mesh->numberOfCells(),
        PUGS_LAMBDA(CellId cell_id) { delta_xj[cell_id] = l2Norm(pk[cell_id](xj[cell_id]) - a[0]); });

      REQUIRE(max(delta_xj) == Catch::Approx(0).margin(1E-14));

      DiscreteFunctionP0<double> delta(mesh_v);
      parallel_for(
        mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) {
          const double x0 = 0.2 * Vj[cell_id];
          const double y0 = 0.3 * Vj[cell_id];
          delta[cell_id]                                     //
            = l2Norm(pk[cell_id](xj[cell_id] + R2{x0, y0})   //
                     - (a[0]                                 //
                        + x0 * a[1]                          //
                        + x0 * x0 * a[2]                     //
                        + x0 * x0 * x0 * a[3]                //
                        + y0 * a[4]                          //
                        + y0 * x0 * a[5]                     //
                        + y0 * x0 * x0 * a[6]                //
                        + y0 * y0 * a[7]                     //
                        + y0 * y0 * x0 * a[8]                //
                        + y0 * y0 * y0 * a[9]));
        });

      REQUIRE(max(delta) == Catch::Approx(0.).margin(1E-14));
    }

    SECTION("3D")
    {
      constexpr size_t Dimension = 3;

      using R2 = TinyVector<2>;
      using R3 = TinyVector<3>;

      const std::shared_ptr mesh_v = MeshDataBaseForTests::get().hybrid3DMesh();
      const std::shared_ptr mesh   = mesh_v->get<Mesh<Dimension>>();

      auto xj = MeshDataManager::instance().getMeshData(*mesh).xj();
      auto Vj = MeshDataManager::instance().getMeshData(*mesh).Vj();

      const size_t degree = 3;

      const std::vector<R2> a = {R2{+9.2, +7.2}, R2{+0.0, -2.7}, R2{+4.8, -5.8},  R2{+6.0, +3.6}, R2{-7.0, -9.7},
                                 R2{+5.3, -1.2}, R2{-1.2, +3.7}, R2{-0.4, +6.1},  R2{+8.4, +9.5}, R2{-9.7, +3.3},
                                 R2{+0.5, +4.2}, R2{+3.8, +3.3}, R2{+8.0, -10.0}, R2{-5.1, -4.1}, R2{+2.6, -2.5},
                                 R2{-3.4, -2.7}, R2{+0.7, +4.9}, R2{+6.0, +6.4},  R2{+3.5, +5.0}, R2{+1.7, +4.8}};

      DiscreteFunctionDPk<Dimension, R2> pk(mesh_v, degree);
      parallel_for(
        mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) {
          auto coefficients = pk.coefficients(cell_id);
          for (size_t i = 0; i < coefficients.size(); ++i) {
            coefficients[i] = a[i];
          }
        });

      DiscreteFunctionP0<double> delta_xj(mesh_v);
      parallel_for(
        mesh->numberOfCells(),
        PUGS_LAMBDA(CellId cell_id) { delta_xj[cell_id] = l2Norm(pk[cell_id](xj[cell_id]) - a[0]); });

      REQUIRE(max(delta_xj) == Catch::Approx(0.).margin(1E-14));

      DiscreteFunctionP0<double> delta(mesh_v);
      parallel_for(
        mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) {
          const double x0 = +0.5 * Vj[cell_id];
          const double y0 = +0.3 * Vj[cell_id];
          const double z0 = -0.4 * Vj[cell_id];

          delta[cell_id] = l2Norm(pk[cell_id](xj[cell_id] + R3{x0, y0, z0})   //
                                  - (a[0]                                     //
                                     + x0 * a[1]                              //
                                     + x0 * x0 * a[2]                         //
                                     + x0 * x0 * x0 * a[3]                    //
                                     + y0 * a[4]                              //
                                     + y0 * x0 * a[5]                         //
                                     + y0 * x0 * x0 * a[6]                    //
                                     + y0 * y0 * a[7]                         //
                                     + y0 * y0 * x0 * a[8]                    //
                                     + y0 * y0 * y0 * a[9]                    //
                                     + z0 * a[10]                             //
                                     + z0 * x0 * a[11]                        //
                                     + z0 * x0 * x0 * a[12]                   //
                                     + z0 * y0 * a[13]                        //
                                     + z0 * y0 * x0 * a[14]                   //
                                     + z0 * y0 * y0 * a[15]                   //
                                     + z0 * z0 * a[16]                        //
                                     + z0 * z0 * x0 * a[17]                   //
                                     + z0 * z0 * y0 * a[18]                   //
                                     + z0 * z0 * z0 * a[19]                   //
                                     ));
        });

      REQUIRE(max(delta) == Catch::Approx(0.).margin(1E-14));
    }
  }

  SECTION("R^d1xd2 data")
  {
    SECTION("1D")
    {
      constexpr size_t Dimension = 1;

      using R1   = TinyVector<1>;
      using R2x3 = TinyMatrix<2, 3>;

      const std::shared_ptr mesh_v = MeshDataBaseForTests::get().cartesian1DMesh();
      const std::shared_ptr mesh   = mesh_v->get<Mesh<Dimension>>();

      auto xj = MeshDataManager::instance().getMeshData(*mesh).xj();
      auto Vj = MeshDataManager::instance().getMeshData(*mesh).Vj();

      const size_t degree = 4;

      DiscreteFunctionDPk<Dimension, R2x3> pk(mesh_v, degree);

      std::vector<R2x3> A = {R2x3{+1.0, +2.0, +3.0, +4.0, +5.0, +6.0},   //
                             R2x3{-1.2, -2.3, +7.2, +8.4, -5.0, +0.7},   //
                             R2x3{-2.1, -3.3, -2.7, -3.4, -0.5, -2.7},   //
                             R2x3{+6.2, -2.9, +3.1, -2.6, +1.5, +2.1},   //
                             R2x3{-2.6, -2.2, +4.2, -1.7, +8.5, -1.4}};

      for (CellId cell_id = 0; cell_id < mesh->numberOfCells(); ++cell_id) {
        auto coefficients = pk.coefficients(cell_id);
        for (size_t i = 0; i < coefficients.size(); ++i) {
          coefficients[i] = A[i];
        }
      };

      DiscreteFunctionP0<double> delta_xj(mesh_v);
      parallel_for(
        mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) {
          const R2x3 Diff = pk[cell_id](xj[cell_id]) - A[0];

          delta_xj[cell_id] = l2Norm((transpose(Diff) * Diff) * TinyVector<3>(1, 1, 1));
        });

      REQUIRE(max(delta_xj) == Catch::Approx(0).margin(1E-14));

      DiscreteFunctionP0<double> delta(mesh_v);
      parallel_for(
        mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) {
          const double x0 = 0.3 * Vj[cell_id];
          const R2x3 Diff = pk[cell_id](xj[cell_id] + R1{x0}) -
                            (A[0] + x0 * A[1] + x0 * x0 * A[2] + x0 * x0 * x0 * A[3] + x0 * x0 * x0 * x0 * A[4]);

          delta[cell_id] = l2Norm((transpose(Diff) * Diff) * TinyVector<3>(1, 1, 1));
        });

      REQUIRE(max(delta) == Catch::Approx(0.).margin(1E-14));
    }

    SECTION("2D")
    {
      constexpr size_t Dimension = 2;

      using R2   = TinyVector<2>;
      using R2x3 = TinyMatrix<2, 3>;

      const std::shared_ptr mesh_v = MeshDataBaseForTests::get().cartesian2DMesh();
      const std::shared_ptr mesh   = mesh_v->get<Mesh<Dimension>>();

      auto xj = MeshDataManager::instance().getMeshData(*mesh).xj();
      auto Vj = MeshDataManager::instance().getMeshData(*mesh).Vj();

      const size_t degree = 3;

      DiscreteFunctionDPk<Dimension, R2x3> pk(mesh_v, degree);

      std::vector<R2x3> A = {R2x3{+1.0, +2.0, +3.0, +4.0, +5.0, +6.0},   //
                             R2x3{-1.2, -2.3, +7.2, +8.4, -5.0, +0.7},   //
                             R2x3{-2.1, -3.3, -2.7, -3.4, -0.5, -2.7},   //
                             R2x3{+6.2, -2.9, +3.1, -2.6, +1.5, +2.1},   //
                             R2x3{-2.6, -2.2, +4.2, -1.7, +8.5, -1.4},   //
                             R2x3{+1.7, +3.1, +3.0, +0.4, +3.4, +4.3},   //
                             R2x3{+2.5, +2.3, +4.7, -8.7, -5.0, +2.4},   //
                             R2x3{-3.6, -1.3, -1.3, -4.1, -0.5, -6.2},   //
                             R2x3{+6.4, -2.9, +2.3, -6.1, +1.5, -1.9},   //
                             R2x3{-7.3, -3.2, +4.1, +2.7, +8.5, -6.9}};

      for (CellId cell_id = 0; cell_id < mesh->numberOfCells(); ++cell_id) {
        auto coefficients = pk.coefficients(cell_id);
        for (size_t i = 0; i < coefficients.size(); ++i) {
          coefficients[i] = A[i];
        }
      };

      DiscreteFunctionP0<double> delta_xj(mesh_v);
      parallel_for(
        mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) {
          const R2x3 Diff   = pk[cell_id](xj[cell_id]) - A[0];
          delta_xj[cell_id] = l2Norm((transpose(Diff) * Diff) * TinyVector<3>(1, 1, 1));
        });

      REQUIRE(max(delta_xj) == Catch::Approx(0).margin(1E-14));

      DiscreteFunctionP0<double> delta(mesh_v);
      parallel_for(
        mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) {
          const double x0 = 0.2 * Vj[cell_id];
          const double y0 = 0.3 * Vj[cell_id];
          const R2x3 Diff = pk[cell_id](xj[cell_id] + R2{x0, y0})   //
                            - (A[0]                                 //
                               + x0 * A[1]                          //
                               + x0 * x0 * A[2]                     //
                               + x0 * x0 * x0 * A[3]                //
                               + y0 * A[4]                          //
                               + y0 * x0 * A[5]                     //
                               + y0 * x0 * x0 * A[6]                //
                               + y0 * y0 * A[7]                     //
                               + y0 * y0 * x0 * A[8]                //
                               + y0 * y0 * y0 * A[9]);

          delta[cell_id] = l2Norm((transpose(Diff) * Diff) * TinyVector<3>(1, 1, 1));
        });

      REQUIRE(max(delta) == Catch::Approx(0.).margin(1E-14));
    }

    SECTION("3D")
    {
      constexpr size_t Dimension = 3;

      using R2x3 = TinyMatrix<2, 3>;
      using R3   = TinyVector<3>;

      const std::shared_ptr mesh_v = MeshDataBaseForTests::get().hybrid3DMesh();
      const std::shared_ptr mesh   = mesh_v->get<Mesh<Dimension>>();

      auto xj = MeshDataManager::instance().getMeshData(*mesh).xj();
      auto Vj = MeshDataManager::instance().getMeshData(*mesh).Vj();

      const size_t degree = 3;

      const std::vector<R2x3> A = {R2x3{-3.1, +0.7, -4.2, -6.6, -2.3, +7.2}, R2x3{-5.1, -4.5, +9.6, +0.8, -3.2, +3.7},
                                   R2x3{-2.4, +3.8, -2.5, +0.7, -6.5, -7.7}, R2x3{+3.2, +7.6, +7.8, -4.1, +1.6, +9.9},
                                   R2x3{-0.8, +6.1, -2.4, -1.1, -4.1, +0.6}, R2x3{+6.3, +1.0, +0.0, -9.4, +4.4, +1.2},
                                   R2x3{-1.1, +9.8, +10., -4.9, -2.4, -4.3}, R2x3{+6.9, +4.2, +8.8, +2.0, +3.4, +6.4},
                                   R2x3{+6.5, -6.3, +7.1, +8.7, -1.9, -9.7}, R2x3{+3.3, +1.5, -8.2, +8.1, +2.9, +3.3},
                                   R2x3{+6.9, -9.8, -2.5, -6.8, +0.9, +9.4}, R2x3{+4.3, +3.0, +2.2, -8.8, +3.0, -3.2},
                                   R2x3{-4.7, +9.0, +2.7, -0.3, -8.1, -8.6}, R2x3{+1.0, +1.7, -3.9, +7.8, -1.2, +5.6},
                                   R2x3{-2.1, +6.1, -8.7, +4.3, -1.9, -9.3}, R2x3{-0.3, -8.3, -9.7, +9.4, -9.7, +3.8},
                                   R2x3{+9.2, +7.1, +9.1, -9.1, +6.8, -5.2}, R2x3{-9.1, +4.8, +5.3, +9.4, -1.2, -9.2},
                                   R2x3{+1.3, -8.7, -1.2, +2.7, -1.8, -1.6}, R2x3{+1.0, -9.7, +1.0, +9.2, -0.1, -4.9}};

      DiscreteFunctionDPk<Dimension, R2x3> pk(mesh_v, degree);
      parallel_for(
        mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) {
          auto coefficients = pk.coefficients(cell_id);
          for (size_t i = 0; i < coefficients.size(); ++i) {
            coefficients[i] = A[i];
          }
        });

      DiscreteFunctionP0<double> delta_xj(mesh_v);
      parallel_for(
        mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) {
          const R2x3 Diff   = pk[cell_id](xj[cell_id]) - A[0];
          delta_xj[cell_id] = l2Norm(transpose(Diff) * Diff * TinyVector<3>{1, 1, 1});
        });

      REQUIRE(max(delta_xj) == Catch::Approx(0.).margin(1E-14));

      DiscreteFunctionP0<double> delta(mesh_v);
      parallel_for(
        mesh->numberOfCells(), PUGS_LAMBDA(CellId cell_id) {
          const double x0 = +0.5 * Vj[cell_id];
          const double y0 = +0.3 * Vj[cell_id];
          const double z0 = -0.4 * Vj[cell_id];

          const R2x3 Diff = pk[cell_id](xj[cell_id] + R3{x0, y0, z0})   //
                            - (A[0]                                     //
                               + x0 * A[1]                              //
                               + x0 * x0 * A[2]                         //
                               + x0 * x0 * x0 * A[3]                    //
                               + y0 * A[4]                              //
                               + y0 * x0 * A[5]                         //
                               + y0 * x0 * x0 * A[6]                    //
                               + y0 * y0 * A[7]                         //
                               + y0 * y0 * x0 * A[8]                    //
                               + y0 * y0 * y0 * A[9]                    //
                               + z0 * A[10]                             //
                               + z0 * x0 * A[11]                        //
                               + z0 * x0 * x0 * A[12]                   //
                               + z0 * y0 * A[13]                        //
                               + z0 * y0 * x0 * A[14]                   //
                               + z0 * y0 * y0 * A[15]                   //
                               + z0 * z0 * A[16]                        //
                               + z0 * z0 * x0 * A[17]                   //
                               + z0 * z0 * y0 * A[18]                   //
                               + z0 * z0 * z0 * A[19]                   //
                              );

          delta[cell_id] = l2Norm(transpose(Diff) * Diff * TinyVector<3>{1, 1, 1});
        });

      REQUIRE(max(delta) == Catch::Approx(0.).margin(1E-14));
    }
  }

  SECTION("degree and polynomial basis size")
  {
    SECTION("1D")
    {
      for (size_t degree = 0; degree < 10; ++degree) {
        size_t polynomial_basis_dimension =
          PolynomialCenteredCanonicalBasisView<1, double>::dimensionFromDegree(degree);
        REQUIRE(polynomial_basis_dimension == degree + 1);
        REQUIRE(PolynomialCenteredCanonicalBasisView<1, double>::degreeFromDimension(polynomial_basis_dimension) ==
                degree);
      }
    }

    SECTION("2D")
    {
      for (size_t degree = 0; degree < 10; ++degree) {
        size_t polynomial_basis_dimension =
          PolynomialCenteredCanonicalBasisView<2, double>::dimensionFromDegree(degree);
        REQUIRE(2 * polynomial_basis_dimension == (degree + 1) * (degree + 2));
        REQUIRE(PolynomialCenteredCanonicalBasisView<2, double>::degreeFromDimension(polynomial_basis_dimension) ==
                degree);
      }

      REQUIRE_THROWS_WITH((PolynomialCenteredCanonicalBasisView<2, double>::degreeFromDimension(2)),
                          "error: incorrect polynomial basis dimension");
      REQUIRE_THROWS_WITH((PolynomialCenteredCanonicalBasisView<2, double>::degreeFromDimension(4)),
                          "error: incorrect polynomial basis dimension");
      REQUIRE_THROWS_WITH((PolynomialCenteredCanonicalBasisView<2, double>::degreeFromDimension(5)),
                          "error: incorrect polynomial basis dimension");
      REQUIRE_THROWS_WITH((PolynomialCenteredCanonicalBasisView<2, double>::degreeFromDimension(7)),
                          "error: incorrect polynomial basis dimension");
    }
  }
}
