#include <catch2/catch_approx.hpp>
#include <catch2/catch_test_macros.hpp>

#include <mesh/DualMeshManager.hpp>
#include <mesh/ItemArrayUtils.hpp>
#include <mesh/ItemValue.hpp>
#include <mesh/ItemValueUtils.hpp>
#include <mesh/Mesh.hpp>
#include <mesh/MeshData.hpp>
#include <mesh/MeshDataManager.hpp>

#include <MeshDataBaseForTests.hpp>

// clazy:excludeall=non-pod-global-static

TEST_CASE("TestVolumes3D", "[scheme]")
{
  SECTION("3D")
  {
    using R3 = TinyVector<3>;

    auto hybrid_mesh = MeshDataBaseForTests::get().hybrid3DMesh();
    /*
    auto f = [](const R3& X) -> double {
               const double x = X[0];
               const double y = X[1];
               const double z = X[2];
               return x * x + 2 * x * y + 3 * y * y + 2 * z * z - z + 1;
             };
    */
    std::vector<std::pair<std::string, decltype(hybrid_mesh)>> mesh_list;
    mesh_list.push_back(std::make_pair("hybrid mesh", hybrid_mesh));
    mesh_list.push_back(std::make_pair("diamond mesh", DualMeshManager::instance().getDiamondDualMesh(hybrid_mesh)));

    for (const auto& mesh_info : mesh_list) {
      auto mesh_name = mesh_info.first;
      auto mesh      = mesh_info.second->get<Mesh<3>>();

      auto& mesh_data = MeshDataManager::instance().getMeshData(*mesh);

      auto Cjr = mesh_data.Cjr();
      auto Cje = mesh_data.Cje();
      auto Cjf = mesh_data.Cjf();

      auto xr = mesh->xr();
      auto Vj = mesh_data.Vj();

      auto xl = mesh_data.xl();   // Les centres des faces
      auto xe = mesh_data.xe();   // Les centres des aretes
      auto Nl = mesh_data.Nl();   // Normale aux faces .. au "centre" des faces
      auto nl = mesh_data.nl();   // Normale unite aux faces .. au "centre" des faces

      NodeValuePerFace<const R3> Nlr = mesh_data.Nlr();

      auto cell_to_node_connectivity = mesh->connectivity().cellToNodeMatrix();
      auto cell_to_face_connectivity = mesh->connectivity().cellToFaceMatrix();

      const auto& cell_face_is_reversed = mesh->connectivity().cellFaceIsReversed();

      auto cell_to_edge_connectivity = mesh->connectivity().cellToEdgeMatrix();
      auto edge_to_node_connectivity = mesh->connectivity().edgeToNodeMatrix();
      auto edge_to_face_connectivity = mesh->connectivity().edgeToFaceMatrix();
      auto face_to_node_connectivity = mesh->connectivity().faceToNodeMatrix();

      FaceValuePerCell<R3> face_normal_per_cell(mesh->connectivity());
      face_normal_per_cell.fill(zero);
      EdgeValuePerCell<R3> edge_normal_per_cell(mesh->connectivity());
      edge_normal_per_cell.fill(zero);
      CellValue<double> tabErrVolCjr(mesh->connectivity());
      CellValue<double> tabErrVolCjf(mesh->connectivity());
      CellValue<double> tabErrVolCje(mesh->connectivity());
      tabErrVolCjr.fill(0.);
      tabErrVolCjf.fill(0.);
      tabErrVolCje.fill(0.);

      CellValue<double> tabErrStokesCjr(mesh->connectivity());
      CellValue<double> tabErrStokesCjf(mesh->connectivity());
      CellValue<double> tabErrStokesCje(mesh->connectivity());
      tabErrStokesCjr.fill(0.);
      tabErrStokesCjf.fill(0.);
      tabErrStokesCje.fill(0.);

      NodeValue<double> tabErrConservation_r_Cjr(mesh->connectivity());
      FaceValue<double> tabErrConservation_f_Cjf(mesh->connectivity());
      EdgeValue<double> tabErrConservation_e_Cje(mesh->connectivity());
      tabErrConservation_r_Cjr.fill(0.);
      tabErrConservation_f_Cjf.fill(0.);
      tabErrConservation_e_Cje.fill(0.);

      parallel_for(
        mesh->numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
          // ********************************
          //  Travail sur les nodes
          // ********************************

          auto cell_nodes = cell_to_node_connectivity[cell_id];
          double SumCjrXr = 0;
          R3 SumCjr       = R3{0, 0, 0};
          for (size_t i_node = 0; i_node < cell_nodes.size(); ++i_node) {
            const NodeId node_id = cell_nodes[i_node];
            SumCjrXr += dot(Cjr(cell_id, i_node), xr[node_id]);
            SumCjr += Cjr(cell_id, i_node);
          }
          SumCjrXr /= 3.;
          tabErrVolCjr[cell_id]    = fabs(Vj[cell_id] - SumCjrXr);
          tabErrStokesCjr[cell_id] = l2Norm(SumCjr);

          // ********************************
          //  Travail sur les faces
          // ********************************

          auto cell_faces              = cell_to_face_connectivity[cell_id];
          const auto& face_is_reversed = cell_face_is_reversed.itemArray(cell_id);

          double SumCjrXf      = 0;
          R3 SumCjf            = R3{0, 0, 0};
          double SumCjrXf_mesh = 0;
          R3 SumCjf_mesh       = R3{0, 0, 0};
          for (size_t i_face = 0; i_face < cell_faces.size(); ++i_face) {
            const FaceId face_id = cell_faces[i_face];
            SumCjrXf_mesh += dot(Cjf(cell_id, i_face), xl[face_id]);
            SumCjf_mesh += Cjf(cell_id, i_face);

            if (face_is_reversed[i_face]) {
              face_normal_per_cell[cell_id][i_face] = -Nl[face_id];
              SumCjrXf -= dot(Nl[face_id], xl[face_id]);
              SumCjf -= Nl[face_id];
            } else {
              face_normal_per_cell[cell_id][i_face] = Nl[face_id];
              SumCjrXf += dot(Nl[face_id], xl[face_id]);
              SumCjf += Nl[face_id];
            }
          }
          SumCjrXf /= 3.;
          SumCjrXf_mesh /= 3.;
          tabErrVolCjf[cell_id]    = fabs(Vj[cell_id] - SumCjrXf);
          tabErrStokesCjf[cell_id] = l2Norm(SumCjf);

          // ********************************
          //  Travail sur les edges
          // ********************************
          // auto cell_faces = cell_to_face_connectivity[cell_id];
          // const auto& face_is_reversed = cell_face_is_reversed.itemArray(cell_id);
          // const auto& cell_face_is_reversed = mesh->connectivity().cellFaceIsReversed();

          auto cell_edges      = cell_to_edge_connectivity[cell_id];
          double SumCjrXe      = 0;
          R3 SumCje            = zero;
          double SumCjrXe_mesh = 0;
          R3 SumCje_mesh       = zero;

          for (size_t i_edge = 0; i_edge < cell_edges.size(); ++i_edge) {
            const EdgeId edge_id = cell_edges[i_edge];

            SumCjrXe_mesh += dot(Cje(cell_id, i_edge), xe[edge_id]);
            SumCje_mesh += Cje(cell_id, i_edge);

            auto edges_faces = edge_to_face_connectivity[edge_id];
            // auto cell_faces = cell_to_face_connectivity[edge_id];
            auto edges_nodes = edge_to_node_connectivity[edge_id];

            const NodeId& node0_id = edges_nodes[0];
            const NodeId& node1_id = edges_nodes[1];

            int nbFacEdge = 0;
            // On ne garde que si les 2 faces qui sont dans la maille
            for (size_t i_face = 0; i_face < edges_faces.size(); ++i_face) {
              const FaceId face_id = edges_faces[i_face];

              for (size_t i_face_cell = 0; i_face_cell < cell_faces.size(); ++i_face_cell) {
                const FaceId face_id_cell = cell_faces[i_face_cell];

                // La face autour de l arete appartient à la maille courante : On peut faire le calcul avec la face
                if (face_id_cell == face_id) {
                  double sign = (cell_face_is_reversed(cell_id, i_face_cell)) ? -1 : 1;

                  const auto& face_nodes = face_to_node_connectivity[face_id_cell];

                  for (size_t rl = 0; rl < face_nodes.size(); ++rl) {
                    if (node0_id == face_nodes[rl] or node1_id == face_nodes[rl])
                      edge_normal_per_cell(cell_id, i_edge) += .5 * sign * Nlr(face_id_cell, rl);
                  }

                  ++nbFacEdge;
                  break;
                }
              }
            }

            if (nbFacEdge != 2) {
              throw UnexpectedError("invaid number of edges");
            }

            R3 Cje_l = edge_normal_per_cell[cell_id][i_edge];
            // R3 Cje_l=Cje(cell_id,i_edge);//edge_normal_per_cell[cell_id][i_edge];

            SumCjrXe += dot(Cje_l, xe[edge_id]);
            SumCje += Cje_l;
          }
          SumCjrXe /= 3.;
          SumCjrXe_mesh /= 3.;

          tabErrVolCje[cell_id]    = fabs(Vj[cell_id] - SumCjrXe);
          tabErrStokesCje[cell_id] = l2Norm(SumCje);
        });

      std::clog << " Max Error Vol et Stokes " << max(tabErrVolCje) << " / " << max(tabErrStokesCje);
      auto node_to_cell_connectivity = mesh->connectivity().nodeToCellMatrix();
      auto face_to_cell_connectivity = mesh->connectivity().faceToCellMatrix();
      auto edge_to_cell_connectivity = mesh->connectivity().edgeToCellMatrix();

      auto is_boundary_node = mesh->connectivity().isBoundaryNode();
      auto is_boundary_face = mesh->connectivity().isBoundaryFace();
      auto is_boundary_edge = mesh->connectivity().isBoundaryEdge();

      parallel_for(
        mesh->numberOfNodes(), PUGS_LAMBDA(const NodeId node_id) {
          R3 SumCjr = R3{0, 0, 0};
          if (not is_boundary_node[node_id]) {
            for (size_t i_node = 0; i_node < node_to_cell_connectivity[node_id].size(); ++i_node) {
              const CellId cell_id = node_to_cell_connectivity[node_id][i_node];

              for (size_t i_node_cell = 0; i_node_cell < cell_to_node_connectivity[cell_id].size(); ++i_node_cell) {
                const NodeId node_cell_id = cell_to_node_connectivity[cell_id][i_node_cell];
                if (node_id == node_cell_id) {
                  SumCjr += Cjr(cell_id, i_node_cell);
                  break;
                }
              }
            }
          }
          tabErrConservation_r_Cjr[node_id] = l2Norm(SumCjr);
        });

      parallel_for(
        mesh->numberOfFaces(), PUGS_LAMBDA(const FaceId face_id) {
          R3 SumCjf = zero;
          if (not is_boundary_face[face_id]) {
            for (size_t i_face = 0; i_face < face_to_cell_connectivity[face_id].size(); ++i_face) {
              const CellId cell_id = face_to_cell_connectivity[face_id][i_face];

              for (size_t i_face_cell = 0; i_face_cell < cell_to_face_connectivity[cell_id].size(); ++i_face_cell) {
                const FaceId face_cell_id = cell_to_face_connectivity[cell_id][i_face_cell];
                if (face_id == face_cell_id) {
                  SumCjf += face_normal_per_cell[cell_id][i_face_cell];
                  break;
                }
              }
            }
          }
          tabErrConservation_f_Cjf[face_id] = l2Norm(SumCjf);
        });

      parallel_for(
        mesh->numberOfEdges(), PUGS_LAMBDA(const EdgeId edge_id) {
          R3 SumCje = zero;

          if (not is_boundary_edge[edge_id]) {
            for (size_t i_edge = 0; i_edge < edge_to_cell_connectivity[edge_id].size(); ++i_edge) {
              const CellId cell_id = edge_to_cell_connectivity[edge_id][i_edge];

              for (size_t i_edge_cell = 0; i_edge_cell < cell_to_edge_connectivity[cell_id].size(); ++i_edge_cell) {
                const EdgeId edge_cell_id = cell_to_edge_connectivity[cell_id][i_edge_cell];
                if (edge_id == edge_cell_id) {
                  SumCje += edge_normal_per_cell[cell_id][i_edge_cell];
                  // SumCje+=Cje(cell_id,i_edge_cell);//edge_normal_per_cell[cell_id][i_edge_cell];
                  break;
                }
              }
            }
          }
          tabErrConservation_e_Cje[edge_id] = l2Norm(SumCje);
        });

      REQUIRE(sum(tabErrConservation_e_Cje) == Catch::Approx(0).margin(1E-13));
    }
  }
}
